/* CS 758 asst 10

   dfs for finding paths in graphs

   Wheeler Ruml, October 2024
*/


#include <stdlib.h> /* calloc */
#include <math.h> /* ceil */
#include <float.h> /* DBL_MAX */
#include "utils.h"
#include "graph.h"


/**************** bitvectors ***********************

  bits are numbered continuously from leftmost bit 0
 */


typedef unsigned char bitvec;


/* all bits set to 0. */
bitvec *calloc_bits(unsigned int num_bits)
{
  unsigned char *bits = (unsigned char *) calloc((size_t) ceil(((float) num_bits) / 8.),
						 sizeof(unsigned char));
  
  if (! bits) panic("failed to malloc DFS bitvec");
  return bits;
}


unsigned char get_bit(bitvec *vec, unsigned int index)
{
  unsigned char byte = vec[index / 8];
  unsigned char bit = index % 8;

  return (byte >> (7 - bit)) & 1;
}


void set_bit(bitvec *vec, unsigned int index)
{
  unsigned int byte_index = index / 8;
  unsigned char bit_index = index % 8;
  unsigned char byte = vec[byte_index];
  unsigned char stencil = 1 << (7 - bit_index);
  
  vec[byte_index] = byte | stencil;
}


void clear_bit(bitvec *vec, unsigned int index)
{
  unsigned int byte_index = index / 8;
  unsigned char bit_index = index % 8;
  unsigned char byte = vec[byte_index];
  unsigned char stencil = ~(1 << (7 - bit_index));
  
  vec[byte_index] = byte & stencil;
}


/**************** finding a path ***********************/


/* DFS visit to curr.  returns non-zero if goal found.  uses path as stack.  */
int visit(const graph *g, const unsigned int goal,
	   bitvec *visited, path *path, unsigned int curr)
{
  unsigned int arci;
  
  path->vertices[path->len++] = curr;
  if (curr == goal) return TRUE;
  set_bit(visited, curr);
  for (arci = g->arcs_end[curr-1]; arci < g->arcs_end[curr]; arci++) {
    unsigned int dest = g->arcs[arci].dest;

    if (! get_bit(visited, dest)) {
      if (visit(g, goal, visited, path, dest))
	return TRUE;
    }
  }
  path->len--;
  return FALSE;
}


/* DFS traversal of g: visits each reachable vertex at most once,
   stops when it finds a path to goal. */
path *dfs_first(const graph *g, const unsigned int start, const unsigned int goal)
{
  bitvec *visited = calloc_bits(g->num_vertices + 1);
  path *path = alloc_path(g->num_vertices);
  int found = visit(g, goal, visited, path, start);
  
  free(visited);
  if (found) {
    return path;
  } else {
    free_path(path);
    return NULL;
  }
}


/**************** enumeration of all paths ***********************/


/* bookkeeping details for DFS search.  does not own any of the pointed-to memory. */
typedef struct search_info {
  const graph *g;
  unsigned int goal;
  path *best_path;
  float best_path_cost;
} info;

  
/* DFS visit to curr  */
void visit_all(info *i, bitvec *visited, path *curr_path, float cost_so_far, unsigned int curr)
{
  unsigned int arci;
  
  curr_path->vertices[curr_path->len++] = curr;
  if (curr == i->goal) {
    /* cost already checked */
    path_copy(i->best_path, curr_path);
    i->best_path_cost = cost_so_far;
  } else {
    unsigned int *ends = i->g->arcs_end;
    arc *arcs = i->g->arcs;
    
    set_bit(visited, curr);
    for (arci = ends[curr-1]; arci < ends[curr]; arci++) {
      unsigned int dest = arcs[arci].dest;
      
      /* enforce simple path */
      if (! get_bit(visited, dest)) {
	float dest_cost = cost_so_far + arcs[arci].cost;

	/* branch-and-bound */
	if (dest_cost < i->best_path_cost)
	  visit_all(i, visited, curr_path, dest_cost, dest);
      }
    }
    clear_bit(visited, curr);
  }
  curr_path->len--;
}


/* enumerates all paths from start in g.  records shortest path to goal */
path *dfs_shortest(const graph *g, const unsigned int start, const unsigned int goal)
{
  info *i = (info *) safe_malloc(sizeof(info), "failed to malloc dfs info");
  bitvec *visited = calloc_bits(g->num_vertices + 1);
  path *curr_path = alloc_path(g->num_vertices);
  path *best_path = alloc_path(g->num_vertices);

  i->g = g;
  i->goal = goal;
  i->best_path = best_path;
  i->best_path_cost = DBL_MAX;
  visit_all(i, visited, curr_path, 0., start);
  free_path(curr_path);
  free(visited);
  free(i);
  if (best_path->len == 0) {
    free_path(best_path);
    return NULL;
  } else {
    return best_path;
  }
}


/* EOF */
