/* UNH CS 758 assignment 10

   shortest paths in graphs

*/

#include <stdio.h>
#include <stdlib.h>  /* free, EXIT_SUCCESS */
#include <string.h>  /* strcmp. memcpy */
#include "graph.h"
#include "utils.h"


/**************** the graph ***********************/


unsigned int read_unsigned_int(FILE *infile)
{
  unsigned int x;
  size_t objs;

  while ((objs = fread(&x, sizeof(unsigned int), 1, infile)) != 1) {
    if (feof(infile)) panic("unexpected EOF while reading graph");
    if (ferror(infile)) panic("error while reading graph");
  }
  return x;
}


unsigned char read_unsigned_char(FILE *infile)
{
  unsigned char x;
  size_t objs;

  while ((objs = fread(&x, sizeof(unsigned char), 1, infile)) != 1) {
    if (feof(infile)) panic("unexpected EOF while reading graph");
    if (ferror(infile)) panic("error while reading graph");
  }
  return x;
}


float read_float(FILE *infile)
{
  float x;
  size_t objs;

  while ((objs = fread(&x, sizeof(float), 1, infile)) != 1) {
    if (feof(infile)) panic("unexpected EOF while reading graph");
    if (ferror(infile)) panic("error while reading graph");
  }
  return x;
}


void read_vertices(FILE *infile, unsigned int num_v, unsigned int *ends,
		   unsigned int num_a, arc *arcs)
{
  unsigned int vi;
  unsigned int ai = 0;
  
  ends[0] = 0;
  for (vi = 1; vi <= num_v; vi++) {
    unsigned char num_this_arcs;
    unsigned int end_ai;
    
    /* ignore lat and lon */
    read_float(infile);
    read_float(infile);
    num_this_arcs = read_unsigned_char(infile);
    for (end_ai = ai + num_this_arcs; ai < end_ai; ai++) {
      unsigned int dest = read_unsigned_int(infile);
      
      if ((dest < 1) || (dest > num_v)) panic("arc dest out of range");
      arcs[ai].dest = dest;
      arcs[ai].cost = read_float(infile);
    }
    ends[vi] = end_ai;
  }
}


graph *read_graph(FILE *infile)
{
  unsigned int num_v = read_unsigned_int(infile);
  unsigned int num_a = read_unsigned_int(infile);
  unsigned int *ends = (unsigned int *) safe_malloc(sizeof(unsigned int)*(num_v+1),
						    "mallocing arc_ends failed");
  arc *arcs = (arc *) safe_malloc(sizeof(arc)*(num_a), "mallocing arcs failed");
  graph *g = (graph *) safe_malloc(sizeof(graph), "mallocing graph failed");
  
  if (sizeof(unsigned int) != 4)
    fprintf(stderr, "WARNING: unsigned int is %lu bytes on the this platform, not 4.\n",
	    sizeof(unsigned int));
  
  read_vertices(infile, num_v, ends, num_a, arcs);
  g->num_vertices = num_v;
  g->arcs_end = ends;
  g->arcs = arcs;
  return g;
}


void free_graph(graph *g)
{
  free(g->arcs_end);
  free(g->arcs);
  free(g);
}


/**************** paths ***********************/


path *alloc_path(unsigned int capacity)
{
  unsigned int *vs = (unsigned int *) safe_malloc(sizeof(unsigned int) * capacity,
						  "failed to malloc path vertices");
  path *p = (path *) safe_malloc(sizeof(path), "failed to malloc path");
  
  p->len = 0;
  p->vertices = vs;
  return p;
}


/* assumes that the capacity of dest is large enough! */
void path_copy(path *dest, path *src)
{
  unsigned int n = src->len;
  
  memcpy(dest->vertices, src->vertices, sizeof(unsigned int) * n);
  dest->len = n;
}


void fprint_path(FILE *out, double time, path *p)
{
  fprintf(out, "%f", time);
  if (p == NULL) {
    fprintf(out, " No path found.\n");
  } else {
    int i;
    
    for (i = 0; i < p->len; i++) {
      fprintf(out, " %u", p->vertices[i]);
    }
    fprintf(out, "\n");
  }
  fflush(NULL);
}


void free_path(path *p)
{
  if (p != NULL) {
    free(p->vertices);
    free(p);
  }
}


/**************** queries ***********************/


/* sets start to 0 on EOF */
void read_start_goal_pair(FILE *in, graph *g, unsigned int *start, unsigned int *goal)
{
  int result = fscanf(in, " %u %u", start, goal);
  if (result == EOF) {
    *start = 0;
  } else if (result != 2) {
    panic("failed to scanf start/goal pair");
  } else if (*start == 0) {
    /* did not get EOF so this is real error*/
    panic("please don't specify dummy vertex 0 as start");
  } else if ((*start > g->num_vertices) || (*goal > g->num_vertices)) {
    panic("specified vertex index is larger than those in graph");
  }
}


void solve_queries(FILE *in, FILE *out, solver f, graph *g)
{
  while (TRUE) {
    unsigned int start, goal;
    path *p;
    double before, after, duration;
    
    read_start_goal_pair(in, g, &start, &goal);
    if (start == 0) break;
    fprintf(stderr, "Finding path...");
    fflush(stderr);
    before = get_current_seconds();
    p = (*f)(g, start, goal);
    after = get_current_seconds();
    duration = after-before;
    fprintf(stderr, "%f secs.\n", duration);
    fprint_path(out, duration, p);
    free_path(p);
  }
}


/**************** main ***********************/


void usage(char *const argv[])
{
  fprintf(stderr, "usage: %s [alg-name]\n\n", argv[0]);
  fprintf(stderr, "Reads graph and queries from stdin, writes paths to stdout, diagnostics to stderr.\n");
  exit(EXIT_FAILURE);
}


int main(int argc, char *const argv[])
{
  graph *g;
  /* avoids uninitialized var warning from gcc */
  solver *s = &dijkstra;

  if (argc != 2) usage(argv);
  if (strcmp(argv[1], "dijkstra") == 0) {
    s = &dijkstra;
  } else if (strcmp(argv[1], "dfs_first") == 0) {
    s = &dfs_first;
  } else if (strcmp(argv[1], "dfs_shortest") == 0) {
    s = &dfs_shortest;
  } else {
    panic("unrecognized algorithm: try dijkstra, dfs_first, or dfs_shortest");
  }
  fprintf(stderr, "Reading graph...\n");
  g = read_graph(stdin);
  solve_queries(stdin, stdout, s, g);
  free_graph(g);
  return EXIT_SUCCESS;
}


/* EOF */
